import ast
import typing
from flake8_django_on_delete_comment import __version__ as version

from flake8_django_on_delete_comment.ast_helpers import get_cascade_lines_without_noqa

ErrorType = typing.Tuple[int, int, str, type]


class DjangoOnDeleteCommentChecker:
    name = 'flake8-django-on-delete-comment'
    version = version

    def __init__(self, tree: ast.AST, filename: str):
        self.filename = filename
        self.tree = tree

    def run(self) -> typing.Generator[ErrorType, None, None]:
        if 'migrations' not in self.filename:
            cascade_lines_without_noqa = get_cascade_lines_without_noqa(self.filename, self.tree)
            errors = self.get_on_delete_comment_errors(cascade_lines_without_noqa)
            for error in errors:
                yield (*error, type(self))

    def get_on_delete_comment_errors(
        self,
        cascade_lines_without_noqa: typing.Set[int],
    ) -> typing.List[typing.Tuple[int, int, str]]:
        errors: typing.List[typing.Tuple[int, int, str]] = []
        if cascade_lines_without_noqa:
            for line in cascade_lines_without_noqa:
                errors.append((
                    line,
                    0,
                    'CD001 field needs a valid comment for on_delete',
                ))

        return errors



import typing

from setuptools import setup, find_packages


package_name = 'flake8_django_on_delete_comment'


def get_version() -> typing.Optional[str]:
    with open('flake8_django_on_delete_comment/__init__.py', 'r') as f:
        lines = f.readlines()
    for line in lines:
        if line.startswith('__version__'):
            return line.split('=')[-1].strip().strip("'")


def get_long_description() -> str:
    with open('README.md') as f:
        return f.read()


setup(
    name=package_name,
    description='A flake8 extension to validate django models ForeignKey fields on on_delete CASCADE comment',
    classifiers=[
        'Environment :: Console',
        'Framework :: Flake8',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Documentation',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    python_requires='>=3.6',
    include_package_data=True,
    keywords='flake8 on_delete',
    version=get_version(),
    author='BestDoctor',
    author_email='khkaterine@gmail.com',
    install_requires=['flake8', 'setuptools'],
    entry_points={
        'flake8.extension': [
            'CD001 = flake8_django_on_delete_comment.checker:DjangoOnDeleteCommentChecker',
        ],
    },
    url='https://github.com/best-doctor/flake8-django-on-delete-comment',
    license='MIT',
    py_modules=[package_name],
    zip_safe=False,
)



import ast
import typing
import tokenize

NOQA_FOR_CASCADE_ON_DELETE = '# allowed_cascade'


def get_field_lines_with_on_delete_cascade(ast_tree: ast.AST) -> typing.Set:
    field_lines_with_cascade_on_delete = set()
    lines_with_cascade_attr = set()
    for node in ast.walk(ast_tree):
        if not isinstance(node, ast.Call):
            continue
        for keyword in node.keywords:
            if keyword.arg == 'on_delete' and (
                    hasattr(keyword.value, 'value') and keyword.value.attr == 'CASCADE'):  # type: ignore
                field_lines_with_cascade_on_delete.add(node.lineno)
                lines_with_cascade_attr.add(keyword.value.lineno)
                break

        for child_node in ast.iter_child_nodes(node):
            if isinstance(child_node, ast.Attribute) and (child_node.attr == 'CASCADE') and (
                    node.lineno not in lines_with_cascade_attr):
                field_lines_with_cascade_on_delete.add(node.lineno)
    return field_lines_with_cascade_on_delete


def get_lines_with_cascade_noqa(filepath: str) -> typing.Set:
    with tokenize.open(filepath) as f:
        line_numbers_with_noqa = set()
        for line in tokenize.generate_tokens(f.readline):
            if (tokenize.COMMENT in line) and (NOQA_FOR_CASCADE_ON_DELETE in line.string):
                line_numbers_with_noqa.add(line.start[0])
    return line_numbers_with_noqa


def get_cascade_lines_without_noqa(filepath: str, ast_tree: ast.AST) -> typing.Set:
    line_numbers_with_noqa = get_lines_with_cascade_noqa(filepath)
    lines_with_cascade_attr = get_field_lines_with_on_delete_cascade(ast_tree)

    return lines_with_cascade_attr.difference(line_numbers_with_noqa)
